import torch
from itertools import permutations

def ctmrg(T, d, Dcut, max_iter):
    
    #symmetrize
    T = (T + T.permute(3, 1, 2, 0))/2. 
    T = (T + T.permute(0, 2, 1, 3))/2. 
    T = (T + T.permute(2, 3, 0, 1))/2. 
    T = (T + T.permute(1, 0, 3, 2))/2. 
    
    lnZ = 0.0
    truncation_error = 0.0
    #C = torch.randn(d, d, dtype=T.dtype, device=T.device) #T.sum((0,1))
    #E = torch.randn(d, d, d, dtype=T.dtype, device=T.device)#T.sum(1)
    C = T.sum((0,1))
    E = T.sum(1)
    D = d

    sold = torch.zeros(d, dtype=T.dtype, device=T.device)
    diff = 1E1
    for n in range(max_iter):
        A = torch.einsum('ab,eca,bdg,cdfh->efgh', (C, E, E, T)).contiguous().view(D*d, D*d)
        A = (A+A.t())/2.

        D_new = min(D*d, Dcut)
        U, S, V = torch.svd(A)
        s = S/S.max() 
        truncation_error += S[D_new:].sum()/S.sum() 
        P = U[:, :D_new] # projection operator

        #S, U = torch.symeig(A, eigenvectors=True)
        #truncation_error += 1.-S[-D_new:].sum()/S.sum() 
        #P = U[:, -D_new:] # projection operator

        C = (P.t() @ A @ P) #(D, D)
        C = (C+C.t())/2.

        ET = torch.einsum('ldr,adbc->labrc', (E, T)).contiguous().view(D*d, d, D*d)
        #ET = torch.tensordot(E, T, dims=([1], [1]))
        #ET = ET.permute(0, 2, 3, 1, 4).contiguous().view(D*d, d, D*d)

        E = torch.einsum('li,ldr,rj->idj', (P, ET, P)) #(D_new, d, D_new)
        #E = ( P.t() @ ((ET.view(D*d*d, D*d)@P).view(D*d, d*D_new))).view(D_new,d,D_new)

        E = (E + E.permute(2, 1, 0))/2.

        D = D_new

        C = C/C.norm()
        E = E/E.norm()

        if (s.numel() == sold.numel()):
            diff = (s-sold).norm()
        if (diff < 1E-8):
            break
        sold = s
    print ('ctmrg iterations', n) 
    Z1 = torch.einsum('ab,bcd,fd,gha,hcij,fjk,lg,mil,mk', (C,E,C,E,T,E,C,E,C))
    #CEC = torch.einsum('da,ebd,ce->abc', (C,E,C)).view(1, D**2*d)
    #ETE = torch.einsum('abc,lbdr,mdn->almcrn',(E,T,E)).contiguous().view(D**2*d, D**2*d)
    #ETE = (ETE+ETE.t())/2.
    #Z1 = CEC@ETE@CEC.t()

    Z3 = torch.einsum('ab,bc,cd,da', (C,C,C,C))
    Z2 = torch.einsum('ab,bcd,de,fa,gcf,ge',(C,E,C,C,E,C))
    #print ('    Z1, Z2, Z3:', Z1.item(), Z2.item(), Z3.item())
    lnZ += torch.log(Z1.abs()) + torch.log(Z3.abs()) - 2.*torch.log(Z2.abs())

    return lnZ, truncation_error/n

if __name__=='__main__':
    torch.set_num_threads(1)
    K = torch.tensor([0.44], dtype=torch.float64)
    Dcut = 180
    max_iter = 1000

    c = torch.sqrt(torch.cosh(K)/2.)
    s = torch.sqrt(torch.sinh(K)/2.)
    M = torch.stack([torch.cat([c+s, c-s]), torch.cat([c-s, c+s])])
    T = torch.einsum('ai,aj,ak,al->ijkl', (M, M, M, M)) 

    lnZ, error = ctmrg(T, 2, Dcut, max_iter)
    print (lnZ.item(), error)
